import logging

from py_zipkin.storage import get_default_tracer
from py_zipkin.util import _should_sample
from py_zipkin.util import create_attrs_for_span
from py_zipkin.util import generate_random_64bit_string
from py_zipkin.util import ZipkinAttrs

log = logging.getLogger(__name__)


def _parse_single_header(b3_header):
    """
    Parse out and return the data necessary for generating ZipkinAttrs.

    Returns a dict with the following keys:
        'trace_id':             str or None
        'span_id':              str or None
        'parent_span_id':       str or None
        'sampled_str':          '0', '1', 'd', or None (defer)
    """
    parsed = dict.fromkeys(("trace_id", "span_id", "parent_span_id", "sampled_str"))

    # b3={TraceId}-{SpanId}-{SamplingState}-{ParentSpanId}
    #      (last 2 fields optional)
    # OR
    # b3={SamplingState}
    bits = b3_header.split("-")

    # Handle the lone-sampling-decision case:
    if len(bits) == 1:
        if bits[0] in ("0", "1", "d"):
            parsed["sampled_str"] = bits[0]
            return parsed
        raise ValueError("Invalid sample-only value: %r" % bits[0])
    if len(bits) > 4:
        # Too many segments
        raise ValueError("Too many segments in b3 header: %r" % b3_header)
    parsed["trace_id"] = bits[0]
    if not parsed["trace_id"]:
        raise ValueError("Bad or missing TraceId")
    parsed["span_id"] = bits[1]
    if not parsed["span_id"]:
        raise ValueError("Bad or missing SpanId")
    if len(bits) > 3:
        parsed["parent_span_id"] = bits[3]
        if not parsed["parent_span_id"]:
            raise ValueError("Got empty ParentSpanId")
    if len(bits) > 2:
        # Empty-string means "missing" which means "Defer"
        if bits[2]:
            parsed["sampled_str"] = bits[2]
            if parsed["sampled_str"] not in ("0", "1", "d"):
                raise ValueError("Bad SampledState: %r" % parsed["sampled_str"])
    return parsed


def _parse_multi_header(headers):
    """
    Parse out and return the data necessary for generating ZipkinAttrs.

    Returns a dict with the following keys:
        'trace_id':             str or None
        'span_id':              str or None
        'parent_span_id':       str or None
        'sampled_str':          '0', '1', 'd', or None (defer)
    """
    parsed = {
        "trace_id": headers.get("X-B3-TraceId", None),
        "span_id": headers.get("X-B3-SpanId", None),
        "parent_span_id": headers.get("X-B3-ParentSpanId", None),
        "sampled_str": headers.get("X-B3-Sampled", None),
    }
    # Normalize X-B3-Flags and X-B3-Sampled to None, '0', '1', or 'd'
    if headers.get("X-B3-Flags") == "1":
        parsed["sampled_str"] = "d"
    if parsed["sampled_str"] == "true":
        parsed["sampled_str"] = "1"
    elif parsed["sampled_str"] == "false":
        parsed["sampled_str"] = "0"
    if parsed["sampled_str"] not in (None, "1", "0", "d"):
        raise ValueError("Got invalid X-B3-Sampled: %s" % parsed["sampled_str"])
    for k in ("trace_id", "span_id", "parent_span_id"):
        if parsed[k] == "":
            raise ValueError("Got empty-string %r" % k)
    if parsed["trace_id"] and not parsed["span_id"]:
        raise ValueError("Got X-B3-TraceId but not X-B3-SpanId")
    elif parsed["span_id"] and not parsed["trace_id"]:
        raise ValueError("Got X-B3-SpanId but not X-B3-TraceId")

    # Handle the common case of no headers at all
    if not parsed["trace_id"] and not parsed["sampled_str"]:
        raise ValueError()  # won't trigger a log message

    return parsed


def extract_zipkin_attrs_from_headers(
    headers, sample_rate=100.0, use_128bit_trace_id=False
):
    """
    Implements extraction of B3 headers per:
        https://github.com/openzipkin/b3-propagation

    The input headers can be any dict-like container that supports "in"
    membership test and a .get() method that accepts a default value.

    Returns a ZipkinAttrs instance or None
    """
    try:
        if "b3" in headers:
            parsed = _parse_single_header(headers["b3"])
        else:
            parsed = _parse_multi_header(headers)
    except ValueError as e:
        if str(e):
            log.warning(e)
        return None

    # Handle the lone-sampling-decision case:
    if not parsed["trace_id"]:
        if parsed["sampled_str"] in ("1", "d"):
            sample_rate = 100.0
        else:
            sample_rate = 0.0
        attrs = create_attrs_for_span(
            sample_rate=sample_rate,
            use_128bit_trace_id=use_128bit_trace_id,
            flags="1" if parsed["sampled_str"] == "d" else "0",
        )
        return attrs

    # Handle any sampling decision, including if it was deferred
    if parsed["sampled_str"]:
        # We have 1==Accept, 0==Deny, d==Debug
        if parsed["sampled_str"] in ("1", "d"):
            is_sampled = True
        else:
            is_sampled = False
    else:
        # sample flag missing; means "Defer" and we're responsible for
        # rolling fresh dice
        is_sampled = _should_sample(sample_rate)

    return ZipkinAttrs(
        parsed["trace_id"],
        parsed["span_id"],
        parsed["parent_span_id"],
        "1" if parsed["sampled_str"] == "d" else "0",
        is_sampled,
    )


def create_http_headers(
    context_stack=None,
    tracer=None,
    new_span_id=False,
):
    """
    Generate the headers for a new zipkin span.

    .. note::

        If the method is not called from within a zipkin_trace context,
        empty dict will be returned back.

    :returns: dict containing (X-B3-TraceId, X-B3-SpanId, X-B3-ParentSpanId,
                X-B3-Flags and X-B3-Sampled) keys OR an empty dict.
    """
    if tracer:
        zipkin_attrs = tracer.get_zipkin_attrs()
    elif context_stack:
        zipkin_attrs = context_stack.get()
    else:
        zipkin_attrs = get_default_tracer().get_zipkin_attrs()

    # If zipkin_attrs is still not set then we're not in a trace context
    if not zipkin_attrs:
        return {}

    if new_span_id:
        span_id = generate_random_64bit_string()
        parent_span_id = zipkin_attrs.span_id
    else:
        span_id = zipkin_attrs.span_id
        parent_span_id = zipkin_attrs.parent_span_id

    return {
        "X-B3-TraceId": zipkin_attrs.trace_id,
        "X-B3-SpanId": span_id,
        "X-B3-ParentSpanId": parent_span_id,
        "X-B3-Flags": "0",
        "X-B3-Sampled": "1" if zipkin_attrs.is_sampled else "0",
    }



import json

from py_zipkin import thrift
from py_zipkin.encoding import protobuf
from py_zipkin.encoding._types import Encoding
from py_zipkin.encoding._types import Kind
from py_zipkin.exception import ZipkinError


def get_encoder(encoding):
    """Creates encoder object for the given encoding.

    :param encoding: desired output encoding protocol.
    :type encoding: Encoding
    :return: corresponding IEncoder object
    :rtype: IEncoder
    """
    if encoding == Encoding.V1_THRIFT:
        return _V1ThriftEncoder()
    if encoding == Encoding.V1_JSON:
        return _V1JSONEncoder()
    if encoding == Encoding.V2_JSON:
        return _V2JSONEncoder()
    if encoding == Encoding.V2_PROTO3:
        return _V2ProtobufEncoder()
    raise ZipkinError(f"Unknown encoding: {encoding}")


class IEncoder:
    """Encoder interface."""

    def fits(self, current_count, current_size, max_size, new_span):
        """Returns whether the new span will fit in the list.

        :param current_count: number of spans already in the list.
        :type current_count: int
        :param current_size: sum of the sizes of all the spans already in the list.
        :type current_size: int
        :param max_size: max supported transport payload size.
        :type max_size: int
        :param new_span: encoded span object that we want to add the the list.
        :type new_span: str or bytes
        :return: True if the new span can be added to the list, False otherwise.
        :rtype: bool
        """
        raise NotImplementedError()

    def encode_span(self, span):
        """Encodes a single span.

        :param span: Span object representing the span.
        :type span: Span
        :return: encoded span.
        :rtype: str or bytes
        """
        raise NotImplementedError()

    def encode_queue(self, queue):
        """Encodes a list of pre-encoded spans.

        :param queue: list of encoded spans.
        :type queue: list
        :return: encoded list, type depends on the encoding.
        :rtype: str or bytes
        """
        raise NotImplementedError()


class _V1ThriftEncoder(IEncoder):
    """Thrift encoder for V1 spans."""

    def fits(self, current_count, current_size, max_size, new_span):
        """Checks if the new span fits in the max payload size.

        Thrift lists have a fixed-size header and no delimiters between elements
        so it's easy to compute the list size.
        """
        return thrift.LIST_HEADER_SIZE + current_size + len(new_span) <= max_size

    def encode_remote_endpoint(self, remote_endpoint, kind, binary_annotations):
        thrift_remote_endpoint = thrift.create_endpoint(
            remote_endpoint.port,
            remote_endpoint.service_name,
            remote_endpoint.ipv4,
            remote_endpoint.ipv6,
        )
        if kind == Kind.CLIENT:
            key = thrift.zipkin_core.SERVER_ADDR
        elif kind == Kind.SERVER:
            key = thrift.zipkin_core.CLIENT_ADDR

        binary_annotations.append(
            thrift.create_binary_annotation(
                key=key,
                value=thrift.SERVER_ADDR_VAL,
                annotation_type=thrift.zipkin_core.AnnotationType.BOOL,
                host=thrift_remote_endpoint,
            )
        )

    def encode_span(self, v2_span):
        """Encodes the current span to thrift."""
        span = v2_span.build_v1_span()

        thrift_endpoint = thrift.create_endpoint(
            span.endpoint.port,
            span.endpoint.service_name,
            span.endpoint.ipv4,
            span.endpoint.ipv6,
        )

        thrift_annotations = thrift.annotation_list_builder(
            span.annotations,
            thrift_endpoint,
        )

        thrift_binary_annotations = thrift.binary_annotation_list_builder(
            span.binary_annotations,
            thrift_endpoint,
        )

        # Add sa/ca binary annotations
        if v2_span.remote_endpoint:
            self.encode_remote_endpoint(
                v2_span.remote_endpoint,
                v2_span.kind,
                thrift_binary_annotations,
            )

        thrift_span = thrift.create_span(
            span.id,
            span.parent_id,
            span.trace_id,
            span.name,
            thrift_annotations,
            thrift_binary_annotations,
            span.timestamp,
            span.duration,
        )

        encoded_span = thrift.span_to_bytes(thrift_span)
        return encoded_span

    def encode_queue(self, queue):
        """Converts the queue to a thrift list"""
        return thrift.encode_bytes_list(queue)


class _BaseJSONEncoder(IEncoder):
    """V1 and V2 JSON encoders need many common helper functions"""

    def fits(self, current_count, current_size, max_size, new_span):
        """Checks if the new span fits in the max payload size.

        Json lists only have a 2 bytes overhead from '[]' plus 1 byte from
        ',' between elements
        """
        return 2 + current_count + current_size + len(new_span) <= max_size

    def _create_json_endpoint(self, endpoint, is_v1):
        """Converts an Endpoint to a JSON endpoint dict.

        :param endpoint: endpoint object to convert.
        :type endpoint: Endpoint
        :param is_v1: whether we're serializing a v1 span. This is needed since
            in v1 some fields default to an empty string rather than being
            dropped if they're not set.
        :type is_v1: bool
        :return: dict representing a JSON endpoint.
        :rtype: dict
        """
        json_endpoint = {}

        if endpoint.service_name:
            json_endpoint["serviceName"] = endpoint.service_name
        elif is_v1:
            # serviceName is mandatory in v1
            json_endpoint["serviceName"] = ""
        if endpoint.port and endpoint.port != 0:
            json_endpoint["port"] = endpoint.port
        if endpoint.ipv4 is not None:
            json_endpoint["ipv4"] = endpoint.ipv4
        if endpoint.ipv6 is not None:
            json_endpoint["ipv6"] = endpoint.ipv6

        return json_endpoint

    def encode_queue(self, queue):
        """Concatenates the list to a JSON list"""
        return "[" + ",".join(queue) + "]"


class _V1JSONEncoder(_BaseJSONEncoder):
    """JSON encoder for V1 spans."""

    def encode_remote_endpoint(self, remote_endpoint, kind, binary_annotations):
        json_remote_endpoint = self._create_json_endpoint(remote_endpoint, True)
        if kind == Kind.CLIENT:
            key = "sa"
        elif kind == Kind.SERVER:
            key = "ca"

        binary_annotations.append(
            {"key": key, "value": True, "endpoint": json_remote_endpoint}
        )

    def encode_span(self, v2_span):
        """Encodes a single span to JSON."""
        span = v2_span.build_v1_span()

        json_span = {
            "traceId": span.trace_id,
            "name": span.name,
            "id": span.id,
            "annotations": [],
            "binaryAnnotations": [],
        }

        if span.parent_id:
            json_span["parentId"] = span.parent_id
        if span.timestamp:
            json_span["timestamp"] = int(span.timestamp * 1000000)
        if span.duration:
            json_span["duration"] = int(span.duration * 1000000)

        v1_endpoint = self._create_json_endpoint(span.endpoint, True)

        for key, timestamp in span.annotations.items():
            json_span["annotations"].append(
                {
                    "endpoint": v1_endpoint,
                    "timestamp": int(timestamp * 1000000),
                    "value": key,
                }
            )

        for key, value in span.binary_annotations.items():
            json_span["binaryAnnotations"].append(
                {"key": key, "value": value, "endpoint": v1_endpoint}
            )

        # Add sa/ca binary annotations
        if v2_span.remote_endpoint:
            self.encode_remote_endpoint(
                v2_span.remote_endpoint,
                v2_span.kind,
                json_span["binaryAnnotations"],
            )

        encoded_span = json.dumps(json_span)

        return encoded_span


class _V2JSONEncoder(_BaseJSONEncoder):
    """JSON encoder for V2 spans."""

    def encode_span(self, span):
        """Encodes a single span to JSON."""

        json_span = {
            "traceId": span.trace_id,
            "id": span.span_id,
        }

        if span.name:
            json_span["name"] = span.name
        if span.parent_id:
            json_span["parentId"] = span.parent_id
        if span.timestamp:
            json_span["timestamp"] = int(span.timestamp * 1000000)
        if span.duration:
            json_span["duration"] = int(span.duration * 1000000)
        if span.shared is True:
            json_span["shared"] = True
        if span.kind and span.kind.value is not None:
            json_span["kind"] = span.kind.value
        if span.local_endpoint:
            json_span["localEndpoint"] = self._create_json_endpoint(
                span.local_endpoint,
                False,
            )
        if span.remote_endpoint:
            json_span["remoteEndpoint"] = self._create_json_endpoint(
                span.remote_endpoint,
                False,
            )
        if span.tags and len(span.tags) > 0:
            # Ensure that tags are all strings
            json_span["tags"] = {
                str(key): str(value) for key, value in span.tags.items()
            }

        if span.annotations:
            json_span["annotations"] = [
                {"timestamp": int(timestamp * 1000000), "value": key}
                for key, timestamp in span.annotations.items()
            ]

        encoded_span = json.dumps(json_span)

        return encoded_span


class _V2ProtobufEncoder(IEncoder):
    """Protobuf encoder for V2 spans."""

    def fits(self, current_count, current_size, max_size, new_span):
        """Checks if the new span fits in the max payload size."""
        return current_size + len(new_span) <= max_size

    def encode_span(self, span):
        """Encodes a single span to protobuf."""
        if not protobuf.installed():
            raise ZipkinError(
                "protobuf encoding requires installing the protobuf's extra "
                "requirements. Use py-zipkin[protobuf] in your requirements.txt."
            )

        pb_span = protobuf.create_protobuf_span(span)
        return protobuf.encode_pb_list([pb_span])

    def encode_queue(self, queue):
        """Concatenates the list to a protobuf list and encodes it to bytes"""
        return b"".join(queue)



import os
import time

from py_zipkin import Kind
from py_zipkin.encoding._encoders import get_encoder
from py_zipkin.encoding._helpers import copy_endpoint_with_new_service_name
from py_zipkin.encoding._helpers import Span
from py_zipkin.exception import ZipkinError
from py_zipkin.transport import BaseTransportHandler


LOGGING_END_KEY = "py_zipkin.logging_end"


class ZipkinLoggingContext:
    """A logging context specific to a Zipkin trace. If the trace is sampled,
    the logging context sends serialized Zipkin spans to a transport_handler.
    The logging context sends root "server" or "client" span, as well as all
    local child spans collected within this context.

    This class should only be used by the main `zipkin_span` entrypoint.
    """

    def __init__(
        self,
        zipkin_attrs,
        endpoint,
        span_name,
        transport_handler,
        report_root_timestamp,
        get_tracer,
        service_name,
        binary_annotations=None,
        add_logging_annotation=False,
        client_context=False,
        max_span_batch_size=None,
        firehose_handler=None,
        encoding=None,
        annotations=None,
    ):
        self.zipkin_attrs = zipkin_attrs
        self.endpoint = endpoint
        self.span_name = span_name
        self.transport_handler = transport_handler
        self.response_status_code = 0
        self._get_tracer = get_tracer
        self.service_name = service_name
        self.report_root_timestamp = report_root_timestamp
        self.tags = binary_annotations or {}
        self.add_logging_annotation = add_logging_annotation
        self.client_context = client_context
        self.max_span_batch_size = max_span_batch_size
        self.firehose_handler = firehose_handler
        self.annotations = annotations or {}

        self.remote_endpoint = None
        self.encoder = get_encoder(encoding)

    def start(self):
        """Actions to be taken before request is handled."""

        # Record the start timestamp.
        self.start_timestamp = time.time()
        return self

    def stop(self):
        """Actions to be taken post request handling."""

        self.emit_spans()

    def emit_spans(self):
        """Main function to log all the annotations stored during the entire
        request. This is done if the request is sampled and the response was
        a success. It also logs the service (`ss` and `sr`) or the client
        ('cs' and 'cr') annotations.
        """

        # FIXME: Should have a single aggregate handler
        if self.firehose_handler:
            # FIXME: We need to allow different batching settings per handler
            self._emit_spans_with_span_sender(
                ZipkinBatchSender(
                    self.firehose_handler, self.max_span_batch_size, self.encoder
                )
            )

        if not self.zipkin_attrs.is_sampled:
            self._get_tracer().clear()
            return

        span_sender = ZipkinBatchSender(
            self.transport_handler, self.max_span_batch_size, self.encoder
        )

        self._emit_spans_with_span_sender(span_sender)
        self._get_tracer().clear()

    def _emit_spans_with_span_sender(self, span_sender):
        with span_sender:
            end_timestamp = time.time()

            # Collect, annotate, and log client spans from the logging handler
            for span in self._get_tracer()._span_storage:
                span.local_endpoint = copy_endpoint_with_new_service_name(
                    self.endpoint,
                    span.local_endpoint.service_name,
                )

                span_sender.add_span(span)

            if self.add_logging_annotation:
                self.annotations[LOGGING_END_KEY] = time.time()

            span_sender.add_span(
                Span(
                    trace_id=self.zipkin_attrs.trace_id,
                    name=self.span_name,
                    parent_id=self.zipkin_attrs.parent_span_id,
                    span_id=self.zipkin_attrs.span_id,
                    kind=Kind.CLIENT if self.client_context else Kind.SERVER,
                    timestamp=self.start_timestamp,
                    duration=end_timestamp - self.start_timestamp,
                    local_endpoint=self.endpoint,
                    remote_endpoint=self.remote_endpoint,
                    shared=not self.report_root_timestamp,
                    annotations=self.annotations,
                    tags=self.tags,
                )
            )


class ZipkinBatchSender:

    MAX_PORTION_SIZE = 100

    def __init__(self, transport_handler, max_portion_size, encoder):
        self.transport_handler = transport_handler
        self.max_portion_size = max_portion_size or self.MAX_PORTION_SIZE
        self.encoder = encoder

        if isinstance(self.transport_handler, BaseTransportHandler):
            self.max_payload_bytes = self.transport_handler.get_max_payload_bytes()
        else:
            self.max_payload_bytes = None

    def __enter__(self):
        self._reset_queue()
        return self

    def __exit__(self, _exc_type, _exc_value, _exc_traceback):
        if any((_exc_type, _exc_value, _exc_traceback)):
            filename = os.path.split(_exc_traceback.tb_frame.f_code.co_filename)[1]
            error = "({}:{}) {}: {}".format(
                filename,
                _exc_traceback.tb_lineno,
                _exc_type.__name__,
                _exc_value,
            )
            raise ZipkinError(error)
        else:
            self.flush()

    def _reset_queue(self):
        self.queue = []
        self.current_size = 0

    def add_span(self, internal_span):
        encoded_span = self.encoder.encode_span(internal_span)

        # If we've already reached the max batch size or the new span doesn't
        # fit in max_payload_bytes, send what we've collected until now and
        # start a new batch.
        is_over_size_limit = (
            self.max_payload_bytes is not None
            and not self.encoder.fits(
                current_count=len(self.queue),
                current_size=self.current_size,
                max_size=self.max_payload_bytes,
                new_span=encoded_span,
            )
        )
        is_over_portion_limit = len(self.queue) >= self.max_portion_size
        if is_over_size_limit or is_over_portion_limit:
            self.flush()

        self.queue.append(encoded_span)
        self.current_size += len(encoded_span)

    def flush(self):
        if self.transport_handler and len(self.queue) > 0:

            message = self.encoder.encode_queue(self.queue)
            self.transport_handler(message)
        self._reset_queue()



import socket
import struct

from py_zipkin.encoding._types import Kind
from py_zipkin.util import unsigned_hex_to_signed_int

try:
    from py_zipkin.encoding.protobuf import zipkin_pb2
except ImportError:  # pragma: no cover
    zipkin_pb2 = None


def installed():
    """Checks whether the protobud library is installed and can be used.

    :return: True if everything's fine, False otherwise
    :rtype: bool
    """
    return zipkin_pb2 is not None


def encode_pb_list(pb_spans):
    """Encode list of protobuf Spans to binary.

    :param pb_spans: list of protobuf Spans.
    :type pb_spans: list of zipkin_pb2.Span
    :return: encoded list.
    :rtype: bytes
    """
    pb_list = zipkin_pb2.ListOfSpans()
    pb_list.spans.extend(pb_spans)
    return pb_list.SerializeToString()


def create_protobuf_span(span):
    """Converts a py_zipkin Span in a protobuf Span.

    :param span: py_zipkin Span to convert.
    :type span: py_zipkin.encoding.Span
    :return: protobuf's Span
    :rtype: zipkin_pb2.Span
    """

    # Protobuf's composite types (i.e. Span's local_endpoint) are immutable.
    # So we can't create a zipkin_pb2.Span here and then set the appropriate
    # fields since `pb_span.local_endpoint = zipkin_pb2.Endpoint` fails.
    # Instead we just create the kwargs and pass them in to the Span constructor.
    pb_kwargs = {}

    pb_kwargs["trace_id"] = _hex_to_bytes(span.trace_id)

    if span.parent_id:
        pb_kwargs["parent_id"] = _hex_to_bytes(span.parent_id)

    pb_kwargs["id"] = _hex_to_bytes(span.span_id)

    pb_kind = _get_protobuf_kind(span.kind)
    if pb_kind:
        pb_kwargs["kind"] = pb_kind

    if span.name:
        pb_kwargs["name"] = span.name
    if span.timestamp:
        pb_kwargs["timestamp"] = int(span.timestamp * 1000 * 1000)
    if span.duration:
        pb_kwargs["duration"] = int(span.duration * 1000 * 1000)

    if span.local_endpoint:
        pb_kwargs["local_endpoint"] = _convert_endpoint(span.local_endpoint)

    if span.remote_endpoint:
        pb_kwargs["remote_endpoint"] = _convert_endpoint(span.remote_endpoint)

    if len(span.annotations) > 0:
        pb_kwargs["annotations"] = _convert_annotations(span.annotations)

    if len(span.tags) > 0:
        pb_kwargs["tags"] = span.tags

    if span.debug:
        pb_kwargs["debug"] = span.debug

    if span.shared:
        pb_kwargs["shared"] = span.shared

    return zipkin_pb2.Span(**pb_kwargs)


def _hex_to_bytes(hex_id):
    """Encodes to hexadecimal ids to big-endian binary.

    :param hex_id: hexadecimal id to encode.
    :type hex_id: str
    :return: binary representation.
    :type: bytes
    """
    if len(hex_id) <= 16:
        int_id = unsigned_hex_to_signed_int(hex_id)
        return struct.pack(">q", int_id)
    else:
        # There's no 16-bytes encoding in Python's struct. So we convert the
        # id as 2 64 bit ids and then concatenate the result.

        # NOTE: we count 16 chars from the right (:-16) rather than the left so
        # that ids with less than 32 chars will be correctly pre-padded with 0s.
        high_id = unsigned_hex_to_signed_int(hex_id[:-16])
        high_bin = struct.pack(">q", high_id)

        low_id = unsigned_hex_to_signed_int(hex_id[-16:])
        low_bin = struct.pack(">q", low_id)

        return high_bin + low_bin


def _get_protobuf_kind(kind):
    """Converts py_zipkin's Kind to Protobuf's Kind.

    :param kind: py_zipkin's Kind.
    :type kind: py_zipkin.Kind
    :return: correcponding protobuf's kind value.
    :rtype: zipkin_pb2.Span.Kind
    """
    if kind == Kind.CLIENT:
        return zipkin_pb2.Span.CLIENT
    elif kind == Kind.SERVER:
        return zipkin_pb2.Span.SERVER
    elif kind == Kind.PRODUCER:
        return zipkin_pb2.Span.PRODUCER
    elif kind == Kind.CONSUMER:
        return zipkin_pb2.Span.CONSUMER
    return None


def _convert_endpoint(endpoint):
    """Converts py_zipkin's Endpoint to Protobuf's Endpoint.

    :param endpoint: py_zipkins' endpoint to convert.
    :type endpoint: py_zipkin.encoding.Endpoint
    :return: corresponding protobuf's endpoint.
    :rtype: zipkin_pb2.Endpoint
    """
    pb_endpoint = zipkin_pb2.Endpoint()

    if endpoint.service_name:
        pb_endpoint.service_name = endpoint.service_name
    if endpoint.port and endpoint.port != 0:
        pb_endpoint.port = endpoint.port
    if endpoint.ipv4:
        pb_endpoint.ipv4 = socket.inet_pton(socket.AF_INET, endpoint.ipv4)
    if endpoint.ipv6:
        pb_endpoint.ipv6 = socket.inet_pton(socket.AF_INET6, endpoint.ipv6)

    return pb_endpoint


def _convert_annotations(annotations):
    """Converts py_zipkin's annotations dict to protobuf.

    :param annotations: annotations dict.
    :type annotations: dict
    :return: corresponding protobuf's list of annotations.
    :rtype: list
    """
    pb_annotations = []
    for value, ts in annotations.items():
        pb_annotations.append(
            zipkin_pb2.Annotation(timestamp=int(ts * 1000 * 1000), value=value)
        )
    return pb_annotations



from py_zipkin.transport import BaseTransportHandler


class MockTransportHandler(BaseTransportHandler):
    """Mock transport for use in tests.

    It doesn't emit anything and just stores the generated spans in memory.
    To check what has been emitted you can use `get_payloads` and get back
    the list of encoded spans that were emitted.
    To use it:

    .. code-block:: python

        transport = MockTransportHandler()
        with zipkin.zipkin_span(
            service_name='test_service_name',
            span_name='test_span_name',
            transport_handler=transport,
            sample_rate=100.0,
            encoding=Encoding.V2_JSON,
        ):
            do_something()

        spans = transport.get_payloads()
        assert len(spans) == 1
        decoded_spans = json.loads(spans[0])
        assert decoded_spans == [{}]
    """

    def __init__(self, max_payload_bytes=None):
        """Creates a new MockTransportHandler.

        :param max_payload_bytes: max payload size in bytes. You often don't
            need to set this in tests unless you want to test what happens
            when your spans are bigger than the maximum payload size.
        :type max_payload_bytes: int
        """
        self.max_payload_bytes = max_payload_bytes
        self.payloads = []

    def send(self, payload):
        """Overrides the real send method. Should not be called directly."""
        self.payloads.append(payload)
        return payload

    def get_max_payload_bytes(self):
        """Overrides the real method. Should not be called directly."""
        return self.max_payload_bytes

    def get_payloads(self):
        """Returns the encoded spans that were sent.

        Spans are batched before being sent, so most of the time the returned
        list will contain only one element. Each element is gonna be an encoded
        list of spans.
        """
        return self.payloads



import socket
from collections import namedtuple
from collections import OrderedDict

from py_zipkin.encoding._types import Kind
from py_zipkin.exception import ZipkinError


Endpoint = namedtuple("Endpoint", ["service_name", "ipv4", "ipv6", "port"])


_V1Span = namedtuple(
    "V1Span",
    [
        "trace_id",
        "name",
        "parent_id",
        "id",
        "timestamp",
        "duration",
        "endpoint",
        "annotations",
        "binary_annotations",
        "remote_endpoint",
    ],
)


class Span:
    """Internal V2 Span representation."""

    def __init__(
        self,
        trace_id,
        name,
        parent_id,
        span_id,
        kind,
        timestamp,
        duration,
        local_endpoint=None,
        remote_endpoint=None,
        debug=False,
        shared=False,
        annotations=None,
        tags=None,
    ):
        """Creates a new Span.

        :param trace_id: Trace id.
        :type trace_id: str
        :param name: Name of the span.
        :type name: str
        :param parent_id: Parent span id.
        :type parent_id: str
        :param span_id: Span id.
        :type span_id: str
        :param kind: Span type (client, server, local, etc...)
        :type kind: Kind
        :param timestamp: start timestamp in seconds.
        :type timestamp: float
        :param duration: span duration in seconds.
        :type duration: float
        :param local_endpoint: the host that recorded this span.
        :type local_endpoint: Endpoint
        :param remote_endpoint: the remote service.
        :type remote_endpoint: Endpoint
        :param debug: True is a request to store this span even if it
            overrides sampling policy.
        :type debug: bool
        :param shared: True if we are contributing to a span started by
            another tracer (ex on a different host).
        :type shared: bool
        :param annotations: Optional dict of str -> timestamp annotations.
        :type annotations: dict
        :param tags: Optional dict of str -> str span tags.
        :type tags: dict
        """
        self.trace_id = trace_id
        self.name = name
        self.parent_id = parent_id
        self.span_id = span_id
        self.kind = kind
        self.timestamp = timestamp
        self.duration = duration
        self.local_endpoint = local_endpoint
        self.remote_endpoint = remote_endpoint
        self.debug = debug
        self.shared = shared
        self.annotations = annotations or {}
        self.tags = tags or {}

        if not isinstance(kind, Kind):
            raise ZipkinError(f"Invalid kind value {kind}. Must be of type Kind.")

        if local_endpoint and not isinstance(local_endpoint, Endpoint):
            raise ZipkinError("Invalid local_endpoint value. Must be of type Endpoint.")

        if remote_endpoint and not isinstance(remote_endpoint, Endpoint):
            raise ZipkinError(
                "Invalid remote_endpoint value. Must be of type Endpoint."
            )

    def __eq__(self, other):  # pragma: no cover
        """Compare function to help assert span1 == span2 in py3"""
        return self.__dict__ == other.__dict__

    def __cmp__(self, other):  # pragma: no cover
        """Compare function to help assert span1 == span2 in py2"""
        return self.__dict__ == other.__dict__

    def __str__(self):  # pragma: no cover
        """Compare function to nicely print Span rather than just the pointer"""
        return str(self.__dict__)

    def build_v1_span(self):
        """Builds and returns a V1 Span.

        :return: newly generated _V1Span
        :rtype: _V1Span
        """
        annotations = OrderedDict([])
        if self.kind == Kind.CLIENT:
            annotations["cs"] = self.timestamp
            annotations["cr"] = self.timestamp + self.duration
        elif self.kind == Kind.SERVER:
            annotations["sr"] = self.timestamp
            annotations["ss"] = self.timestamp + self.duration
        elif self.kind == Kind.PRODUCER:
            annotations["ms"] = self.timestamp
        elif self.kind == Kind.CONSUMER:
            annotations["mr"] = self.timestamp

        # Add user-defined annotations. We write them in annotations
        # instead of the opposite so that user annotations will override
        # any automatically generated annotation.
        annotations.update(self.annotations)

        return _V1Span(
            trace_id=self.trace_id,
            name=self.name,
            parent_id=self.parent_id,
            id=self.span_id,
            timestamp=self.timestamp if self.shared is False else None,
            duration=self.duration if self.shared is False else None,
            endpoint=self.local_endpoint,
            annotations=annotations,
            binary_annotations=self.tags,
            remote_endpoint=self.remote_endpoint,
        )


def create_endpoint(port=None, service_name=None, host=None, use_defaults=True):
    """Creates a new Endpoint object.

    :param port: TCP/UDP port. Defaults to 0.
    :type port: int
    :param service_name: service name as a str. Defaults to 'unknown'.
    :type service_name: str
    :param host: ipv4 or ipv6 address of the host. Defaults to the
    current host ip.
    :type host: str
    :param use_defaults: whether to use defaults.
    :type use_defaults: bool
    :returns: zipkin Endpoint object
    """
    if use_defaults:
        if port is None:
            port = 0
        if service_name is None:
            service_name = "unknown"
        if host is None:
            try:
                host = socket.gethostbyname(socket.gethostname())
            except socket.gaierror:
                host = "127.0.0.1"

    ipv4 = None
    ipv6 = None

    if host:
        # Check ipv4 or ipv6.
        try:
            socket.inet_pton(socket.AF_INET, host)
            ipv4 = host
        except OSError:
            # If it's not an ipv4 address, maybe it's ipv6.
            try:
                socket.inet_pton(socket.AF_INET6, host)
                ipv6 = host
            except OSError:
                # If it's neither ipv4 or ipv6, leave both ip addresses unset.
                pass

    return Endpoint(ipv4=ipv4, ipv6=ipv6, port=port, service_name=service_name)


def copy_endpoint_with_new_service_name(endpoint, new_service_name):
    """Creates a copy of a given endpoint with a new service name.

    :param endpoint: existing Endpoint object
    :type endpoint: Endpoint
    :param new_service_name: new service name
    :type new_service_name: str
    :returns: zipkin new Endpoint object
    """
    return Endpoint(
        service_name=new_service_name,
        ipv4=endpoint.ipv4,
        ipv6=endpoint.ipv6,
        port=endpoint.port,
    )



import os
import socket
import struct

import thriftpy2
from thriftpy2.protocol import TBinaryProtocol
from thriftpy2.protocol.binary import write_list_begin
from thriftpy2.thrift import TType
from thriftpy2.transport import TMemoryBuffer

from py_zipkin.util import unsigned_hex_to_signed_int


thrift_filepath = os.path.join(os.path.dirname(__file__), "zipkinCore.thrift")
zipkin_core = thriftpy2.load(thrift_filepath, module_name="zipkinCore_thrift")

SERVER_ADDR_VAL = "\x01"
LIST_HEADER_SIZE = 5  # size in bytes of the encoded list header

dummy_endpoint = zipkin_core.Endpoint()


def create_annotation(timestamp, value, host):
    """
    Create a zipkin annotation object

    :param timestamp: timestamp of when the annotation occured in microseconds
    :param value: name of the annotation, such as 'sr'
    :param host: zipkin endpoint object

    :returns: zipkin annotation object
    """
    return zipkin_core.Annotation(timestamp=timestamp, value=value, host=host)


def create_binary_annotation(key, value, annotation_type, host):
    """
    Create a zipkin binary annotation object

    :param key: name of the annotation, such as 'http.uri'
    :param value: value of the annotation, such as a URI
    :param annotation_type: type of annotation, such as AnnotationType.I32
    :param host: zipkin endpoint object

    :returns: zipkin binary annotation object
    """
    return zipkin_core.BinaryAnnotation(
        key=key,
        value=value,
        annotation_type=annotation_type,
        host=host,
    )


def create_endpoint(port=0, service_name="unknown", ipv4=None, ipv6=None):
    """Create a zipkin Endpoint object.

    An Endpoint object holds information about the network context of a span.

    :param port: int value of the port. Defaults to 0
    :param service_name: service name as a str. Defaults to 'unknown'
    :param ipv4: ipv4 host address
    :param ipv6: ipv6 host address
    :returns: thrift Endpoint object
    """
    ipv4_int = 0
    ipv6_binary = None

    # Convert ip address to network byte order
    if ipv4:
        ipv4_int = struct.unpack("!i", socket.inet_pton(socket.AF_INET, ipv4))[0]

    if ipv6:
        ipv6_binary = socket.inet_pton(socket.AF_INET6, ipv6)

    # Zipkin passes unsigned values in signed types because Thrift has no
    # unsigned types, so we have to convert the value.
    port = struct.unpack("h", struct.pack("H", port))[0]
    return zipkin_core.Endpoint(
        ipv4=ipv4_int,
        ipv6=ipv6_binary,
        port=port,
        service_name=service_name,
    )


def copy_endpoint_with_new_service_name(endpoint, service_name):
    """Copies a copy of a given endpoint with a new service name.
    This should be very fast, on the order of several microseconds.

    :param endpoint: existing zipkin_core.Endpoint object
    :param service_name: str of new service name
    :returns: zipkin Endpoint object
    """
    return zipkin_core.Endpoint(
        ipv4=endpoint.ipv4,
        port=endpoint.port,
        service_name=service_name,
    )


def annotation_list_builder(annotations, host):
    """
    Reformat annotations dict to return list of corresponding zipkin_core objects.

    :param annotations: dict containing key as annotation name,
                        value being timestamp in seconds(float).
    :type host: :class:`zipkin_core.Endpoint`
    :returns: a list of annotation zipkin_core objects
    :rtype: list
    """
    return [
        create_annotation(int(timestamp * 1000000), key, host)
        for key, timestamp in annotations.items()
    ]


def binary_annotation_list_builder(binary_annotations, host):
    """
    Reformat binary annotations dict to return list of zipkin_core objects. The
    value of the binary annotations MUST be in string format.

    :param binary_annotations: dict with key, value being the name and value
                               of the binary annotation being logged.
    :type host: :class:`zipkin_core.Endpoint`
    :returns: a list of binary annotation zipkin_core objects
    :rtype: list
    """
    # TODO: Remove the type hard-coding of STRING to take it as a param option.
    ann_type = zipkin_core.AnnotationType.STRING
    return [
        create_binary_annotation(key, str(value), ann_type, host)
        for key, value in binary_annotations.items()
    ]


def create_span(
    span_id,
    parent_span_id,
    trace_id,
    span_name,
    annotations,
    binary_annotations,
    timestamp_s,
    duration_s,
):
    """Takes a bunch of span attributes and returns a thriftpy2 representation
    of the span. Timestamps passed in are in seconds, they're converted to
    microseconds before thrift encoding.
    """
    # Check if trace_id is 128-bit. If so, record trace_id_high separately.
    trace_id_length = len(trace_id)
    trace_id_high = None
    if trace_id_length > 16:
        assert trace_id_length == 32
        trace_id, trace_id_high = trace_id[16:], trace_id[:16]

    if trace_id_high:
        trace_id_high = unsigned_hex_to_signed_int(trace_id_high)

    span_dict = {
        "trace_id": unsigned_hex_to_signed_int(trace_id),
        "name": span_name,
        "id": unsigned_hex_to_signed_int(span_id),
        "annotations": annotations,
        "binary_annotations": binary_annotations,
        "timestamp": int(timestamp_s * 1000000) if timestamp_s else None,
        "duration": int(duration_s * 1000000) if duration_s else None,
        "trace_id_high": trace_id_high,
    }
    if parent_span_id:
        span_dict["parent_id"] = unsigned_hex_to_signed_int(parent_span_id)
    return zipkin_core.Span(**span_dict)


def span_to_bytes(thrift_span):
    """
    Returns a TBinaryProtocol encoded Thrift span.

    :param thrift_span: thrift object to encode.
    :returns: thrift object in TBinaryProtocol format bytes.
    """
    transport = TMemoryBuffer()
    protocol = TBinaryProtocol(transport)
    thrift_span.write(protocol)

    return bytes(transport.getvalue())


def encode_bytes_list(binary_thrift_obj_list):  # pragma: no cover
    """
    Returns a TBinaryProtocol encoded list of Thrift objects.

    :param binary_thrift_obj_list: list of TBinaryProtocol objects to encode.
    :returns: bynary object representing the encoded list.
    """
    transport = TMemoryBuffer()
    write_list_begin(transport, TType.STRUCT, len(binary_thrift_obj_list))
    for thrift_bin in binary_thrift_obj_list:
        transport.write(thrift_bin)

    return bytes(transport.getvalue())



import random
import struct
import time
from collections import namedtuple


"""
Holds the basic attributes needed to log a zipkin trace

:param trace_id: Unique trace id
:param span_id: Span Id of the current request span
:param parent_span_id: Parent span Id of the current request span
:param flags: stores flags header. Currently unused
:param is_sampled: pre-computed bool whether the trace should be logged
"""
ZipkinAttrs = namedtuple(
    "ZipkinAttrs",
    ["trace_id", "span_id", "parent_span_id", "flags", "is_sampled"],
)


def generate_random_64bit_string():
    """Returns a 64 bit UTF-8 encoded string. In the interests of simplicity,
    this is always cast to a `str` instead of (in py2 land) a unicode string.
    Certain clients (I'm looking at you, Twisted) don't enjoy unicode headers.

    :returns: random 16-character string
    """
    return f"{random.getrandbits(64):016x}"


def generate_random_128bit_string():
    """Returns a 128 bit UTF-8 encoded string. Follows the same conventions
    as generate_random_64bit_string().

    The upper 32 bits are the current time in epoch seconds, and the
    lower 96 bits are random. This allows for AWS X-Ray `interop
    <https://github.com/openzipkin/zipkin/issues/1754>`_

    :returns: 32-character hex string
    """
    t = int(time.time())
    lower_96 = random.getrandbits(96)
    return f"{(t << 96) | lower_96:032x}"


def unsigned_hex_to_signed_int(hex_string):
    """Converts a 64-bit hex string to a signed int value.

    This is due to the fact that Apache Thrift only has signed values.

    Examples:
        '17133d482ba4f605' => 1662740067609015813
        'b6dbb1c2b362bf51' => -5270423489115668655

    :param hex_string: the string representation of a zipkin ID
    :returns: signed int representation
    """
    return struct.unpack("q", struct.pack("Q", int(hex_string, 16)))[0]


def signed_int_to_unsigned_hex(signed_int):
    """Converts a signed int value to a 64-bit hex string.

    Examples:
        1662740067609015813  => '17133d482ba4f605'
        -5270423489115668655 => 'b6dbb1c2b362bf51'

    :param signed_int: an int to convert
    :returns: unsigned hex string
    """
    hex_string = hex(struct.unpack("Q", struct.pack("q", signed_int))[0])[2:]
    if hex_string.endswith("L"):
        return hex_string[:-1]
    return hex_string


def _should_sample(sample_rate):
    if sample_rate == 0.0:
        return False  # save a die roll
    elif sample_rate == 100.0:
        return True  # ditto
    return (random.random() * 100) < sample_rate


def create_attrs_for_span(
    sample_rate=100.0,
    trace_id=None,
    span_id=None,
    use_128bit_trace_id=False,
    flags=None,
):
    """Creates a set of zipkin attributes for a span.

    :param sample_rate: Float between 0.0 and 100.0 to determine sampling rate
    :type sample_rate: float
    :param trace_id: Optional 16-character hex string representing a trace_id.
                    If this is None, a random trace_id will be generated.
    :type trace_id: str
    :param span_id: Optional 16-character hex string representing a span_id.
                    If this is None, a random span_id will be generated.
    :type span_id: str
    :param use_128bit_trace_id: If true, generate 128-bit trace_ids
    :type use_128bit_trace_id: bool
    """
    # Calculate if this trace is sampled based on the sample rate
    if trace_id is None:
        if use_128bit_trace_id:
            trace_id = generate_random_128bit_string()
        else:
            trace_id = generate_random_64bit_string()
    if span_id is None:
        span_id = generate_random_64bit_string()
    is_sampled = _should_sample(sample_rate)

    return ZipkinAttrs(
        trace_id=trace_id,
        span_id=span_id,
        parent_span_id=None,
        flags=flags or "0",
        is_sampled=is_sampled,
    )



from urllib.request import Request
from urllib.request import urlopen

from py_zipkin.encoding import detect_span_version_and_encoding
from py_zipkin.encoding import Encoding


class BaseTransportHandler:
    def get_max_payload_bytes(self):  # pragma: no cover
        """Returns the maximum payload size for this transport.

        Most transports have a maximum packet size that can be sent. For example,
        UDP has a 65507 bytes MTU.
        py_zipkin automatically batches collected spans for performance reasons.
        The batch size is going to be the minimum between `get_max_payload_bytes`
        and `max_span_batch_size` from `zipkin_span`.

        If you don't want to enforce a max payload size, return None.

        :returns: max payload size in bytes or None.
        """
        raise NotImplementedError("get_max_payload_bytes is not implemented")

    def send(self, payload):  # pragma: no cover
        """Sends the encoded payload over the transport.

        :argument payload: encoded list of spans.
        """
        raise NotImplementedError("send is not implemented")

    def __call__(self, payload):
        """Internal wrapper around `send`. Do not override.

        Mostly used to keep backward compatibility with older transports
        implemented as functions. However decoupling the function developers
        override and what's internally called by py_zipkin will allow us to add
        extra logic here in the future without having the users update their
        code every time.
        """
        self.send(payload)


class SimpleHTTPTransport(BaseTransportHandler):
    def __init__(self, address, port):
        """A simple HTTP transport for zipkin.

        This is not production ready (not async, no retries) but
        it's helpful for tests or people trying out py-zipkin.

        .. code-block:: python

            with zipkin_span(
                service_name='my_service',
                span_name='home',
                sample_rate=100,
                transport_handler=SimpleHTTPTransport('localhost', 9411),
                encoding=Encoding.V2_JSON,
            ):
                pass

        :param address: zipkin server address.
        :type address: str
        :param port: zipkin server port.
        :type port: int
        """
        super().__init__()
        self.address = address
        self.port = port

    def get_max_payload_bytes(self):
        return None

    def _get_path_content_type(self, payload):
        """Choose the right api path and content type depending on the encoding.

        This is not something you'd need to do generally when writing your own
        transport since in that case you'd know which encoding you're using.
        Since this is a generic transport, we need to make it compatible with
        any encoding instead.
        """
        encoding = detect_span_version_and_encoding(payload)

        if encoding == Encoding.V1_JSON:
            return "/api/v1/spans", "application/json"
        elif encoding == Encoding.V1_THRIFT:
            return "/api/v1/spans", "application/x-thrift"
        elif encoding == Encoding.V2_JSON:
            return "/api/v2/spans", "application/json"
        elif encoding == Encoding.V2_PROTO3:
            return "/api/v2/spans", "application/x-protobuf"

    def send(self, payload):
        path, content_type = self._get_path_content_type(payload)
        url = f"http://{self.address}:{self.port}{path}"

        req = Request(url, payload, {"Content-Type": content_type})
        response = urlopen(req)

        assert response.getcode() == 202



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: zipkin.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0czipkin.proto\x12\rzipkin.proto3"\xf5\x03\n\x04Span\x12\x10\n\x08trace_id\x18\x01 \x01(\x0c\x12\x11\n\tparent_id\x18\x02 \x01(\x0c\x12\n\n\x02id\x18\x03 \x01(\x0c\x12&\n\x04kind\x18\x04 \x01(\x0e\x32\x18.zipkin.proto3.Span.Kind\x12\x0c\n\x04name\x18\x05 \x01(\t\x12\x11\n\ttimestamp\x18\x06 \x01(\x06\x12\x10\n\x08\x64uration\x18\x07 \x01(\x04\x12/\n\x0elocal_endpoint\x18\x08 \x01(\x0b\x32\x17.zipkin.proto3.Endpoint\x12\x30\n\x0fremote_endpoint\x18\t \x01(\x0b\x32\x17.zipkin.proto3.Endpoint\x12.\n\x0b\x61nnotations\x18\n \x03(\x0b\x32\x19.zipkin.proto3.Annotation\x12+\n\x04tags\x18\x0b \x03(\x0b\x32\x1d.zipkin.proto3.Span.TagsEntry\x12\r\n\x05\x64\x65\x62ug\x18\x0c \x01(\x08\x12\x0e\n\x06shared\x18\r \x01(\x08\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"U\n\x04Kind\x12\x19\n\x15SPAN_KIND_UNSPECIFIED\x10\x00\x12\n\n\x06\x43LIENT\x10\x01\x12\n\n\x06SERVER\x10\x02\x12\x0c\n\x08PRODUCER\x10\x03\x12\x0c\n\x08\x43ONSUMER\x10\x04"J\n\x08\x45ndpoint\x12\x14\n\x0cservice_name\x18\x01 \x01(\t\x12\x0c\n\x04ipv4\x18\x02 \x01(\x0c\x12\x0c\n\x04ipv6\x18\x03 \x01(\x0c\x12\x0c\n\x04port\x18\x04 \x01(\x05".\n\nAnnotation\x12\x11\n\ttimestamp\x18\x01 \x01(\x06\x12\r\n\x05value\x18\x02 \x01(\t"1\n\x0bListOfSpans\x12"\n\x05spans\x18\x01 \x03(\x0b\x32\x13.zipkin.proto3.Span"\x10\n\x0eReportResponse2T\n\x0bSpanService\x12\x45\n\x06Report\x12\x1a.zipkin.proto3.ListOfSpans\x1a\x1d.zipkin.proto3.ReportResponse"\x00\x42\x12\n\x0ezipkin2.proto3P\x01\x62\x06proto3'
)


_SPAN = DESCRIPTOR.message_types_by_name["Span"]
_SPAN_TAGSENTRY = _SPAN.nested_types_by_name["TagsEntry"]
_ENDPOINT = DESCRIPTOR.message_types_by_name["Endpoint"]
_ANNOTATION = DESCRIPTOR.message_types_by_name["Annotation"]
_LISTOFSPANS = DESCRIPTOR.message_types_by_name["ListOfSpans"]
_REPORTRESPONSE = DESCRIPTOR.message_types_by_name["ReportResponse"]
_SPAN_KIND = _SPAN.enum_types_by_name["Kind"]
Span = _reflection.GeneratedProtocolMessageType(
    "Span",
    (_message.Message,),
    {
        "TagsEntry": _reflection.GeneratedProtocolMessageType(
            "TagsEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _SPAN_TAGSENTRY,
                "__module__": "zipkin_pb2"
                # @@protoc_insertion_point(class_scope:zipkin.proto3.Span.TagsEntry)
            },
        ),
        "DESCRIPTOR": _SPAN,
        "__module__": "zipkin_pb2"
        # @@protoc_insertion_point(class_scope:zipkin.proto3.Span)
    },
)
_sym_db.RegisterMessage(Span)
_sym_db.RegisterMessage(Span.TagsEntry)

Endpoint = _reflection.GeneratedProtocolMessageType(
    "Endpoint",
    (_message.Message,),
    {
        "DESCRIPTOR": _ENDPOINT,
        "__module__": "zipkin_pb2"
        # @@protoc_insertion_point(class_scope:zipkin.proto3.Endpoint)
    },
)
_sym_db.RegisterMessage(Endpoint)

Annotation = _reflection.GeneratedProtocolMessageType(
    "Annotation",
    (_message.Message,),
    {
        "DESCRIPTOR": _ANNOTATION,
        "__module__": "zipkin_pb2"
        # @@protoc_insertion_point(class_scope:zipkin.proto3.Annotation)
    },
)
_sym_db.RegisterMessage(Annotation)

ListOfSpans = _reflection.GeneratedProtocolMessageType(
    "ListOfSpans",
    (_message.Message,),
    {
        "DESCRIPTOR": _LISTOFSPANS,
        "__module__": "zipkin_pb2"
        # @@protoc_insertion_point(class_scope:zipkin.proto3.ListOfSpans)
    },
)
_sym_db.RegisterMessage(ListOfSpans)

ReportResponse = _reflection.GeneratedProtocolMessageType(
    "ReportResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _REPORTRESPONSE,
        "__module__": "zipkin_pb2"
        # @@protoc_insertion_point(class_scope:zipkin.proto3.ReportResponse)
    },
)
_sym_db.RegisterMessage(ReportResponse)

_SPANSERVICE = DESCRIPTOR.services_by_name["SpanService"]
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"\n\016zipkin2.proto3P\001"
    _SPAN_TAGSENTRY._options = None
    _SPAN_TAGSENTRY._serialized_options = b"8\001"
    _SPAN._serialized_start = 32
    _SPAN._serialized_end = 533
    _SPAN_TAGSENTRY._serialized_start = 403
    _SPAN_TAGSENTRY._serialized_end = 446
    _SPAN_KIND._serialized_start = 448
    _SPAN_KIND._serialized_end = 533
    _ENDPOINT._serialized_start = 535
    _ENDPOINT._serialized_end = 609
    _ANNOTATION._serialized_start = 611
    _ANNOTATION._serialized_end = 657
    _LISTOFSPANS._serialized_start = 659
    _LISTOFSPANS._serialized_end = 708
    _REPORTRESPONSE._serialized_start = 710
    _REPORTRESPONSE._serialized_end = 726
    _SPANSERVICE._serialized_start = 728
    _SPANSERVICE._serialized_end = 812
# @@protoc_insertion_point(module_scope)



import logging
import threading
from collections import deque

try:  # pragma: no cover
    # Since python 3.7 threadlocal is deprecated in favor of contextvars
    # which also work in asyncio.
    import contextvars

    _contextvars_tracer = contextvars.ContextVar("py_zipkin.Tracer object")
except ImportError:  # pragma: no cover
    # The contextvars module was added in python 3.7
    _contextvars_tracer = None
_thread_local_tracer = threading.local()


log = logging.getLogger("py_zipkin.storage")


def _get_thread_local_tracer():
    """Returns the current tracer from thread-local.

    If there's no current tracer it'll create a new one.
    :returns: current tracer.
    :rtype: Tracer
    """
    if not hasattr(_thread_local_tracer, "tracer"):
        _thread_local_tracer.tracer = Tracer()
    return _thread_local_tracer.tracer


def _set_thread_local_tracer(tracer):
    """Sets the current tracer in thread-local.

    :param tracer: current tracer.
    :type tracer: Tracer
    """
    _thread_local_tracer.tracer = tracer


def _get_contextvars_tracer():  # pragma: no cover
    """Returns the current tracer from contextvars.

    If there's no current tracer it'll create a new one.
    :returns: current tracer.
    :rtype: Tracer
    """
    try:
        return _contextvars_tracer.get()
    except LookupError:
        _contextvars_tracer.set(Tracer())
        return _contextvars_tracer.get()


def _set_contextvars_tracer(tracer):  # pragma: no cover
    """Sets the current tracer in contextvars.

    :param tracer: current tracer.
    :type tracer: Tracer
    """
    _contextvars_tracer.set(tracer)


class Tracer:
    def __init__(self):
        self._is_transport_configured = False
        self._span_storage = SpanStorage()
        self._context_stack = Stack()

    def get_zipkin_attrs(self):
        return self._context_stack.get()

    def push_zipkin_attrs(self, ctx):
        self._context_stack.push(ctx)

    def pop_zipkin_attrs(self):
        return self._context_stack.pop()

    def add_span(self, span):
        self._span_storage.append(span)

    def get_spans(self):
        return self._span_storage

    def clear(self):
        self._span_storage.clear()

    def set_transport_configured(self, configured):
        self._is_transport_configured = configured

    def is_transport_configured(self):
        return self._is_transport_configured

    def zipkin_span(self, *argv, **kwargs):
        from py_zipkin.zipkin import zipkin_span

        kwargs["_tracer"] = self
        return zipkin_span(*argv, **kwargs)

    def copy(self):
        """Return a copy of this instance, but with a deep-copied
        _context_stack.  The use-case is for passing a copy of a Tracer into
        a new thread context.
        """
        the_copy = self.__class__()
        the_copy._is_transport_configured = self._is_transport_configured
        the_copy._span_storage = self._span_storage
        the_copy._context_stack = self._context_stack.copy()
        return the_copy


class Stack:
    """
    Stack is a simple stack class.

    It offers the operations push, pop and get.
    The latter two return None if the stack is empty.

    .. deprecated::
       Use the Tracer interface which offers better multi-threading support.
       Stack will be removed in version 1.0.
    """

    def __init__(self, storage=None):
        if storage is not None:
            log.warning("Passing a storage object to Stack is deprecated.")
            self._storage = storage
        else:
            self._storage = []

    def push(self, item):
        self._storage.append(item)

    def pop(self):
        if self._storage:
            return self._storage.pop()

    def get(self):
        if self._storage:
            return self._storage[-1]

    def copy(self):
        # Return a new Stack() instance with a deep copy of our stack contents
        the_copy = self.__class__()
        the_copy._storage = self._storage[:]
        return the_copy


class ThreadLocalStack(Stack):
    """ThreadLocalStack is variant of Stack that uses a thread local storage.

    The thread local storage is accessed lazily in every method call,
    so the thread that calls the method matters, not the thread that
    instantiated the class.
    Every instance shares the same thread local data.

    .. deprecated::
       Use the Tracer interface which offers better multi-threading support.
       ThreadLocalStack will be removed in version 1.0.
    """

    def __init__(self):
        log.warning(
            "ThreadLocalStack is deprecated. See DEPRECATIONS.rst for"
            "details on how to migrate to using Tracer."
        )

    @property
    def _storage(self):
        return get_default_tracer()._context_stack._storage


class SpanStorage(deque):
    """Stores the list of completed spans ready to be sent.

    .. deprecated::
       Use the Tracer interface which offers better multi-threading support.
       SpanStorage will be removed in version 1.0.
    """

    pass


def default_span_storage():
    log.warning(
        "default_span_storage is deprecated. See DEPRECATIONS.rst for"
        "details on how to migrate to using Tracer."
    )
    return get_default_tracer()._span_storage


def has_default_tracer():
    """Is there a default tracer created already?

    :returns: Is there a default tracer created already?
    :rtype: boolean
    """
    try:
        if _contextvars_tracer and _contextvars_tracer.get():
            return True
    except LookupError:
        pass
    return hasattr(_thread_local_tracer, "tracer")


def get_default_tracer():
    """Return the current default Tracer.

    For now it'll get it from thread-local in Python 2.7 to 3.6 and from
    contextvars since Python 3.7.

    :returns: current default tracer.
    :rtype: Tracer
    """
    if _contextvars_tracer:
        return _get_contextvars_tracer()

    return _get_thread_local_tracer()


def set_default_tracer(tracer):
    """Sets the current default Tracer.

    For now it'll get it from thread-local in Python 2.7 to 3.6 and from
    contextvars since Python 3.7.

    :returns: current default tracer.
    :rtype: Tracer
    """
    if _contextvars_tracer:
        return _set_contextvars_tracer(tracer)

    return _set_thread_local_tracer(tracer)



import json

from py_zipkin.encoding._decoders import get_decoder
from py_zipkin.encoding._encoders import get_encoder
from py_zipkin.encoding._helpers import create_endpoint  # noqa: F401
from py_zipkin.encoding._helpers import Endpoint  # noqa: F401
from py_zipkin.encoding._helpers import Span  # noqa: F401
from py_zipkin.encoding._types import Encoding
from py_zipkin.exception import ZipkinError

_V2_ATTRIBUTES = ["tags", "localEndpoint", "remoteEndpoint", "shared", "kind"]


def detect_span_version_and_encoding(message):
    """Returns the span type and encoding for the message provided.

    The logic in this function is a Python port of
    https://github.com/openzipkin/zipkin/blob/master/zipkin/src/main/java/zipkin/internal/DetectingSpanDecoder.java

    :param message: span to perform operations on.
    :type message: byte array
    :returns: span encoding.
    :rtype: Encoding
    """
    # In case message is sent in as non-bytearray format,
    # safeguard convert to bytearray before handling
    if isinstance(message, str):
        message = message.encode("utf-8")  # pragma: no cover

    if len(message) < 2:
        raise ZipkinError("Invalid span format. Message too short.")

    # Check for binary format
    if message[0] <= 16:
        if message[0] == 10 and message[1:2][0] != 0:
            return Encoding.V2_PROTO3
        return Encoding.V1_THRIFT

    str_msg = message.decode("utf-8")

    # JSON case for list of spans
    if str_msg[0] == "[":
        span_list = json.loads(str_msg)
        if len(span_list) > 0:
            # Assumption: All spans in a list are the same version
            # Logic: Search for identifying fields in all spans, if any span can
            # be strictly identified to a version, return that version.
            # Otherwise, if no spans could be strictly identified, default to V2.
            for span in span_list:
                if any(word in span for word in _V2_ATTRIBUTES):
                    return Encoding.V2_JSON
                elif "binaryAnnotations" in span or (
                    "annotations" in span and "endpoint" in span["annotations"]
                ):
                    return Encoding.V1_JSON
            return Encoding.V2_JSON

    raise ZipkinError("Unknown or unsupported span encoding")


def convert_spans(spans, output_encoding, input_encoding=None):
    """Converts encoded spans to a different encoding.

    param spans: encoded input spans.
    type spans: byte array
    param output_encoding: desired output encoding.
    type output_encoding: Encoding
    param input_encoding: optional input encoding. If this is not specified, it'll
        try to understand the encoding automatically by inspecting the input spans.
    type input_encoding: Encoding
    :returns: encoded spans.
    :rtype: byte array
    """
    if not isinstance(input_encoding, Encoding):
        input_encoding = detect_span_version_and_encoding(message=spans)

    if input_encoding == output_encoding:
        return spans

    decoder = get_decoder(input_encoding)
    encoder = get_encoder(output_encoding)
    decoded_spans = decoder.decode_spans(spans)
    output_spans = []

    # Encode each indivicual span
    for span in decoded_spans:
        output_spans.append(encoder.encode_span(span))

    # Outputs from encoder.encode_span() can be easily concatenated in a list
    return encoder.encode_queue(output_spans)



# Export useful functions and types from private modules.
from py_zipkin.encoding._types import Encoding  # noqa
from py_zipkin.encoding._types import Kind  # noqa
from py_zipkin.storage import get_default_tracer  # noqa
from py_zipkin.storage import Tracer  # noqa



import threading
from functools import partial

from py_zipkin import storage

_orig_Thread_start = threading.Thread.start
_orig_Thread_run = threading.Thread.run


def _Thread_pre_start(self):
    self._orig_tracer = None
    if storage.has_default_tracer():
        self._orig_tracer = storage.get_default_tracer().copy()


def _Thread_wrap_run(self, actual_run_fn):
    # This executes in the new OS thread
    if self._orig_tracer:
        # Inject our copied Tracer into our thread-local-storage
        storage.set_default_tracer(self._orig_tracer)
    try:
        actual_run_fn()
    finally:
        # I think this is probably a good idea for the same reasons the
        # parent class deletes __target, __args, and __kwargs
        if self._orig_tracer:
            del self._orig_tracer


def patch_threading():  # pragma: no cover
    """Monkey-patch threading module to work better with tracing."""

    def _new_start(self):
        _Thread_pre_start(self)
        _orig_Thread_start(self)

    def _new_run(self):
        _Thread_wrap_run(self, partial(_orig_Thread_run, self))

    threading.Thread.start = _new_start
    threading.Thread.run = _new_run


def unpatch_threading():  # pragma: no cover
    threading.Thread.start = _orig_Thread_start
    threading.Thread.run = _orig_Thread_run



import logging

from py_zipkin.storage import get_default_tracer

log = logging.getLogger("py_zipkin.thread_local")


def get_thread_local_zipkin_attrs():
    """A wrapper to return _thread_local.zipkin_attrs

    Returns a list of ZipkinAttrs objects, used for intra-process context
    propagation.

    .. deprecated::
       Use the Tracer interface which offers better multi-threading support.
       get_thread_local_zipkin_attrs will be removed in version 1.0.

    :returns: list that may contain zipkin attribute tuples
    :rtype: list
    """
    log.warning(
        "get_thread_local_zipkin_attrs is deprecated. See DEPRECATIONS.rst"
        " for details on how to migrate to using Tracer."
    )
    return get_default_tracer()._context_stack._storage


def get_thread_local_span_storage():
    """A wrapper to return _thread_local.span_storage

    Returns a SpanStorage object used to temporarily store all spans created in
    the current process. The transport handlers will pull from this storage when
    they emit the spans.

    .. deprecated::
       Use the Tracer interface which offers better multi-threading support.
       get_thread_local_span_storage will be removed in version 1.0.

    :returns: SpanStore object containing all non-root spans.
    :rtype: py_zipkin.storage.SpanStore
    """
    log.warning(
        "get_thread_local_span_storage is deprecated. See DEPRECATIONS.rst"
        " for details on how to migrate to using Tracer."
    )
    return get_default_tracer()._span_storage


def get_zipkin_attrs():
    """Get the topmost level zipkin attributes stored.

    .. deprecated::
       Use the Tracer interface which offers better multi-threading support.
       get_zipkin_attrs will be removed in version 1.0.

    :returns: tuple containing zipkin attrs
    :rtype: :class:`zipkin.ZipkinAttrs`
    """
    from py_zipkin.storage import ThreadLocalStack

    log.warning(
        "get_zipkin_attrs is deprecated. See DEPRECATIONS.rst for"
        "details on how to migrate to using Tracer."
    )
    return ThreadLocalStack().get()


def pop_zipkin_attrs():
    """Pop the topmost level zipkin attributes, if present.

    .. deprecated::
       Use the Tracer interface which offers better multi-threading support.
       pop_zipkin_attrs will be removed in version 1.0.

    :returns: tuple containing zipkin attrs
    :rtype: :class:`zipkin.ZipkinAttrs`
    """
    from py_zipkin.storage import ThreadLocalStack

    log.warning(
        "pop_zipkin_attrs is deprecated. See DEPRECATIONS.rst for"
        "details on how to migrate to using Tracer."
    )
    return ThreadLocalStack().pop()


def push_zipkin_attrs(zipkin_attr):
    """Stores the zipkin attributes to thread local.

    .. deprecated::
       Use the Tracer interface which offers better multi-threading support.
       push_zipkin_attrs will be removed in version 1.0.

    :param zipkin_attr: tuple containing zipkin related attrs
    :type zipkin_attr: :class:`zipkin.ZipkinAttrs`
    """
    from py_zipkin.storage import ThreadLocalStack

    log.warning(
        "push_zipkin_attrs is deprecated. See DEPRECATIONS.rst for"
        "details on how to migrate to using Tracer."
    )
    return ThreadLocalStack().push(zipkin_attr)



from enum import Enum


class Encoding(Enum):
    """Supported output encodings."""

    V1_THRIFT = "V1_THRIFT"
    V1_JSON = "V1_JSON"
    V2_JSON = "V2_JSON"
    V2_PROTO3 = "V2_PROTO3"


class Kind(Enum):
    """Type of Span."""

    CLIENT = "CLIENT"
    SERVER = "SERVER"
    PRODUCER = "PRODUCER"
    CONSUMER = "CONSUMER"
    LOCAL = None



#!/usr/bin/python
import os

from setuptools import find_packages
from setuptools import setup

__version__ = '1.1.0'


def read(f):
    return open(os.path.join(os.path.dirname(__file__), f)).read().strip()


setup(
    name='py_zipkin',
    version=__version__,
    provides=["py_zipkin"],
    author='Yelp, Inc.',
    author_email='opensource+py-zipkin@yelp.com',
    license='Copyright Yelp 2019',
    url="https://github.com/Yelp/py_zipkin",
    description='Library for using Zipkin in Python.',
    long_description='\n\n'.join((read('README.md'), read('CHANGELOG.rst'))),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=('tests*', 'testing*', 'tools*')),
    package_data={'': ['*.thrift']},
    python_requires='>=3.6',
    install_requires=[
        'thriftpy2>=0.4.0',
    ],
    extras_require={
        'protobuf': 'protobuf >= 3.12.4',
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)
